from torch import nn
from transformers import BertModel
from frozen_bert import FrozenBert


# Wrapper Class for BERT - Freezes BERT Layers but otherwise does nothing but wrap bert
class FrozenBert(nn.Module):
    def __init__(self, bert=None, inherit_config=False):
        super(FrozenBert, self).__init__()

        # Define BERT layer
        if bert is None:
            self.bert = BertModel.from_pretrained('bert-base-uncased')
        else:
            self.bert = bert

        # Set config
        if inherit_config:
            self.config = self.bert.config

        # Freeze Bert Parameters
        for param in self.bert.parameters():
            param.requires_grad = False

    def forward(self, input_ids, attention_mask, token_type_ids=None):
        # Return Context Embeddings

        return self.bert(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)


class BertSingleDense(FrozenBert):
    def __init__(self, bert=None, inherit_config=False, name_override=None):
        self.name = "BertSingleDense"
        if name_override is not None:
            if isinstance(name_override, str):
                self.name = name_override

        # Inherit layers from parent
        super().__init__(bert, inherit_config)

        # Add additional layer(s)
        self.dense = nn.Linear(self.bert.config.hidden_size, self.bert.config.vocab_size)  # 1x Dense layer
        self.relu = nn.ReLU()  # ReLU activation function
        self.log_softmax = nn.LogSoftmax(dim=-1)

    def forward(self, input_ids, attention_mask, token_type_ids=None):
        # Obtain context embeddings from frozen BERT model
        frozen_bert_output = super().forward(input_ids, attention_mask, token_type_ids=token_type_ids)

        # Pass context embeddings through unfrozen dense layer
        dense_output = self.dense(frozen_bert_output[0])  # Last hidden layer of BERT - [1] is pooled

        # Obtain results of ReLU activation function applied to dense output
        relu_output = self.relu(dense_output)
        log_softmax_output = self.log_softmax(relu_output)
        return log_softmax_output


class BertBiLSTM(FrozenBert):
    def __init__(self, bert=None, inherit_config=False, name_override=None):
        self.name = "BertBiLSTM"
        if name_override is not None:
            if isinstance(name_override, str):
                self.name = name_override

        super().__init__(bert, inherit_config)

        self.bilstm = nn.LSTM(self.bert.config.hidden_size, self.bert.config.hidden_size, bidirectional=True)
        self.dense = nn.Linear(2 * self.bert.config.hidden_size, self.bert.config.vocab_size)  # Resize Outputs
        self.log_softmax = nn.LogSoftmax(dim=-1)

    def forward(self, input_ids, attention_mask, token_type_ids=None):
        frozen_bert_output = super().forward(input_ids, attention_mask, token_type_ids=token_type_ids)

        lstm_output, _ = self.bilstm(frozen_bert_output[0])
        dense_output = self.dense(lstm_output)

        log_softmax_output = self.log_softmax(dense_output)
        return log_softmax_output


class BertDoubleDense(FrozenBert):
    def __init__(self, bert=None, inherit_config=False, use_softmax=False):
        # Inherit layers from parent
        super().__init__(bert, inherit_config)

        self.name = "BertDoubleDense"

        self.use_softmax = use_softmax

        # Add additional layer(s)
        self.dense1 = nn.Linear(self.bert.config.hidden_size, self.bert.config.hidden_size)  # 1st Dense layer
        self.relu = nn.ReLU()  # ReLU activation function
        self.dense2 = nn.Linear(self.bert.config.hidden_size, self.bert.config.vocab_size)  # 2nd Dense layer
        self.log_softmax = nn.LogSoftmax(dim=-1)

    def forward(self, input_ids, attention_mask, token_type_ids=None):
        frozen_bert_output = super().forward(input_ids, attention_mask, token_type_ids=token_type_ids)

        dense_output1 = self.dense1(frozen_bert_output[0])
        relu_output1 = self.relu(dense_output1)
        dense_output2 = self.dense2(relu_output1)
        relu_output2 = self.relu(dense_output2)

        if self.use_softmax:
            log_softmax_output = self.log_softmax(relu_output2)
            return log_softmax_output
        else:
            return relu_output2


if __name__ == "__main__":
    print("You are running a model class file directly. Go run main.py or equivalent trainer instead!")