from torch.utils.data import Dataset
from transformers import BertTokenizerFast
from datasets import load_dataset
from tqdm import tqdm
import torch


class CNNDailyMailDataset(Dataset):
    def __init__(self, tokenizer=None, split_type='train', split_percentage=None, verbose=0, version='3.0.0',
                 max_input_length=512, preprocess_type='on_get_item'):
        super().__init__()

        # Set dataset's verbose level - used for debugging mostly at the moment
        self.verbose = False

        # TODO Validate
        self.max_input_length = max_input_length

        # Validate tokenizer input
        if tokenizer is None:
            self.tokenizer = BertTokenizerFast.from_pretrained('bert-base-uncased')
        elif isinstance(tokenizer, str):
            self.tokenizer = BertTokenizerFast.from_pretrained(tokenizer)
        else:
            self.tokenizer = tokenizer

        # Validate split_type input
        valid_splits = {'train', 'test', 'validation'}
        if split_type not in valid_splits:
            raise ValueError(f"Incorrect Split type supplied to CNNDailyMailDataset: {split_type}")

        # TODO Do this better
        # Validate split_percentage input
        if split_percentage is not None:
            if isinstance(split_percentage, str):
                split_type = f'{split_type}[:{split_percentage}]'
            elif isinstance(split_percentage, (int, float)):
                if 0 < split_percentage <= 100:
                    split_type = f'{split_type}[:{str(split_percentage)}%]'
            else:
                raise ValueError(f"Incorrect Split Percentage supplied to CNNDailyMailDataset, out of range 0-100")
        else:
            raise ValueError(f"Incorrect typing for Split Percentage supplied to CNNDailyMailDataset,"
                             f" how did you do that? {type(split_percentage).__name__}")

        # Set and validate data preprocessing mode
        valid_preprocess_types = {'on_get_item', 'on_load_dataset'}
        if preprocess_type not in valid_preprocess_types:
            raise ValueError(f"Incorrect preprocess type supplied to CNNDailyMailDataset: {preprocess_type}")
        self.preprocess_type = preprocess_type

        # TODO Add verbose level outputs in more places and adjust to appropriate levels
        # Outputs debug information if verbose level higher than 0
        if verbose > 0:
            print(f"Loading cnn_dailymail dataset {version} with split type: {split_type}")
        if verbose > 1:
            print(f"Loading dataset with preprocess type {self.preprocess_type}")

        # Load dataset
        self.dataset = load_dataset("cnn_dailymail", version, split=split_type)

        # If preprocessing all dataset on load, do so now
        if self.preprocess_type == 'on_load_dataset':
            print(f'Pre-Processing Dataset\n')
            tqdm.pandas()  # Load tqdm for Pandas DataFrame for progress bar
            self.dataset = self.dataset.progress_map(
                lambda data_to_encode: self.encode_with_tokenizer(data_to_encode, tokenizer=tokenizer),
                with_progress_bar=True)
            self.dataset.set_format(type='torch', columns=['input_ids', 'attention_mask', 'labels'])

    # Maybe could be static?
    def encode_with_tokenizer(self, to_encode, tokenizer=None, max_length=None):
        # Set mutable defaults
        if tokenizer is None:
            tokenizer = self.tokenizer
        if max_length is None:
            max_length = self.max_input_length

        # Create encoded article and highlight
        encoded_article = tokenizer.encode_plus(to_encode['article'], truncation=True, padding='max_length',
                                                max_length=max_length)
        encoded_highlights = tokenizer.encode_plus(to_encode['highlights'], truncation=True, padding='max_length',
                                                   max_length=max_length)
        return {
            'input_ids': encoded_article['input_ids'],
            'attention_mask': encoded_article['attention_mask'],
            'labels': encoded_highlights['input_ids']
        }

    def __len__(self):
        return self.dataset.shape[0]

    # TODO Maybe validate preprocess type again? Is there a need?
    def __getitem__(self, index):
        data = self.dataset[index]
        if self.preprocess_type == 'on_get_item':
            article = data['article']
            highlights = data['highlights']
            encoded_article = self.tokenizer.encode_plus(article, truncation=True, padding='max_length',
                                                         max_length=self.max_input_length)
            encoded_highlights = self.tokenizer.encode_plus(highlights, truncation=True, padding='max_length',
                                                            max_length=self.max_input_length)
            return {
                'input_ids': torch.tensor(encoded_article['input_ids'], dtype=torch.long),
                'attention_mask': torch.tensor(encoded_article['attention_mask'], dtype=torch.long),
                'labels': torch.tensor(encoded_highlights['input_ids'], dtype=torch.long)
            }
        # Explicit for now in case I need to add more modes - unlikely
        elif self.preprocess_type == 'on_load_dataset':
            return self.dataset[index]
        else:
            raise NotImplementedError(f"Preprocess type {self.preprocess_type} not yet implemented. Too bad!")
