import torch
from itertools import groupby


def shrink_list(lst):
    groups = groupby(lst)
    result = []

    for key, group in groups:
        count = sum(1 for _ in group)
        if count > 1:
            result.append(f"{key}*{count}")
        else:
            result.append(key)

    return result


# Google's Neural Machine Translation System: Bridging the Gap between Human and Machine Translation
def beam_search_bert(model_outputs, max_length=512, beam_size=3, min_length=None, tokenizer_eos_token_id=None,
                     length_penalty_enabled=True, alpha=0.6):
    device = model_outputs.device

    if len(model_outputs.shape) == 2:
        model_outputs = model_outputs.unsqueeze(0)
    elif len(model_outputs.shape) != 3:
        raise ValueError("model_outputs must have either 2 or 3 dimensions")

    batch_size = model_outputs.shape[0]

    # Initialize arrays for beam stats
    beams = torch.zeros((batch_size, beam_size, max_length), dtype=torch.long, device=device)
    beam_scores = torch.full((batch_size, beam_size), float('-inf'), device=device)
    beam_lengths = torch.zeros((batch_size, beam_size), dtype=torch.long, device=device)

    # Choose the top beam_size tokens for each sequence in the batch
    initial_scores = model_outputs[:, 0, :]
    top_scores, top_tokens = torch.topk(initial_scores, beam_size, dim=-1)

    beams[:, :, 0] = top_tokens
    beam_scores[:, :] = top_scores
    beam_lengths[:, :] = 1

    for i in range(1, model_outputs.shape[1]):
        # calculate the scores for all possible next tokens
        # and keep only the top beam_size tokens for each sequence in the batch
        scores = model_outputs[:, i, :].unsqueeze(1) + beam_scores.unsqueeze(2)
        scores = scores.view(batch_size, -1)

        if min_length is not None and i < min_length and tokenizer_eos_token_id is not None:
            scores[:, tokenizer_eos_token_id] = -1e9

        top_scores, top_indices = torch.topk(scores, beam_size, dim=-1)

        # Convert 2D indices to 3D indices
        prev_beam = top_indices // model_outputs.shape[-1]
        next_token = top_indices % model_outputs.shape[-1]

        beams[:, :, :i] = beams.gather(1, prev_beam.unsqueeze(2).expand(-1, -1, i))
        beams[:, :, i] = next_token

        beam_scores = top_scores
        beam_lengths = beam_lengths.gather(1, prev_beam) + 1

        if length_penalty_enabled:
            # Apply length penalty
            length_penalty = ((5.0 + beam_lengths.float()) / 6.0).pow(alpha)
            beam_scores = beam_scores / length_penalty

    return beams, beam_scores, beam_lengths


def greedy_search(model_outputs):
    # Argmax probability dimension
    max_scores, token_ids = torch.max(model_outputs, dim=-1)
    return token_ids