from pytorch_lightning import LightningModule
from torch import nn
from torch.optim import AdamW
from torchmetrics.functional.text.rouge import rouge_score
from transformers import BertTokenizerFast
import torch
from decode_utils import beam_search_bert, greedy_search, shrink_list


class BertLightning(LightningModule):
    def __init__(self, model, learning_rate=1e-5, tokenizer=None, greedy_decode=True, beam_search_decode=True,
                 beam_width=3, name_override=None):
        super().__init__()
        self.save_hyperparameters()
        self.model = model
        try:
            self.name = self.model.name
        except AttributeError:
            self.name = "BertLightning"
        if name_override:
            self.name = name_override

        self.learning_rate = learning_rate
        self.criterion = nn.NLLLoss()

        self.greedy_decode = greedy_decode
        self.beam_width = beam_width
        self.beam_search_decode = beam_search_decode

        self.validation_step_outputs = []
        self.validation_step_labels = []

        if tokenizer is None:
            tokenizer = BertTokenizerFast.from_pretrained('bert-base-uncased')
        self.tokenizer = tokenizer

    def forward(self, input_ids, attention_mask, token_type_ids=None):
        return self.model(input_ids, attention_mask, token_type_ids)

    def training_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        labels = batch["labels"]

        outputs = self(input_ids, attention_mask)
        loss = self.criterion(outputs.view(-1, self.model.bert.config.vocab_size), labels.view(-1))

        self.log('train_loss', loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        labels = batch["labels"]
        outputs = self(input_ids=input_ids, attention_mask=attention_mask)
        loss = self.criterion(outputs.view(-1, self.model.bert.config.vocab_size), labels.view(-1))
        self.log('val_loss', loss, prog_bar=True)
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        labels = batch["labels"]
        decoded_lables = self.tokenizer.batch_decode(labels, skip_special_tokens=True)
        outputs = self(input_ids=input_ids, attention_mask=attention_mask)
        if self.greedy_decode:
            outputs = greedy_search(outputs)
            decoded_outputs = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)
            scores = rouge_score(decoded_outputs, decoded_lables)
            for key, value in scores.items():
                self.log(f'val_greedy_{key}', value)
        if self.beam_search_decode:
            beams, beam_scores, beam_lengths = beam_search_bert(outputs, length_penalty_enabled=False, beam_size=10)
            best_beam_indices = torch.argmax(beam_scores, dim=-1)
            best_beams = beams[torch.arange(beams.size(0)), best_beam_indices]  # Currently based on highest score
            decoded_outputs = self.tokenizer.batch_decode(best_beams[0], skip_special_tokens=True)
            scores = rouge_score(decoded_outputs, decoded_lables)
            for key, value in scores.items():
                if key == "rouge1_fmeasure" or (key == "val_rouge2_fmeasure" and value != 0.0):
                    self.log(f'val_beam_decode_{key}', value, prog_bar=True)
                else:
                    self.log(f'val_beam_decode_{key}', value)


    @torch.no_grad()
    def test_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        labels = batch["labels"]
        decoded_lables = self.tokenizer.batch_decode(labels, skip_special_tokens=True)
        outputs = self(input_ids=input_ids, attention_mask=attention_mask)
        if self.greedy_decode:
            outputs = greedy_search(outputs)
            decoded_outputs = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)
            scores = rouge_score(decoded_outputs, decoded_lables)
            for key, value in scores.items():
                self.log(f'test_greedy_{key}', value, prog_bar=False)
        if self.beam_search_decode:
            beams, beam_scores, beam_lengths = beam_search_bert(outputs, length_penalty_enabled=False, beam_size=10)
            best_beam_indices = torch.argmax(beam_scores, dim=-1)
            best_beams = beams[torch.arange(beams.size(0)), best_beam_indices]  # Currently based on highest score
            decoded_outputs = self.tokenizer.batch_decode(best_beams[0], skip_special_tokens=True)
            scores = rouge_score(decoded_outputs, decoded_lables)
            for key, value in scores.items():
                if key == "rouge1_fmeasure" or (key == "val_rouge2_fmeasure" and value != 0.0):
                    self.log(f'test_beam_decode_{key}', value, prog_bar=True)
                else:
                    self.log(f'test_beam_decode_{key}', value, prog_bar=False)

    def on_validation_epoch_end(self):
        # Calculate ROUGE score
        scores = rouge_score(self.validation_step_outputs, self.validation_step_labels)

        # Log ROUGE scores
        for key, value in scores.items():
            self.log(f'{self.name}val_{key}', value, prog_bar=True)

    def configure_optimizers(self):
        optimizer = AdamW(self.parameters(), lr=self.learning_rate)
        return optimizer
