from pytorch_lightning.loggers import WandbLogger
from sklearn.model_selection import KFold
from torch.optim import AdamW
from base_models import *
from lightning_models import BertLightning
from torchmetrics.text.rouge import ROUGEScore
from pprint import pprint
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint
from lightning_fabric.loggers import CSVLogger

from torch.utils.data import DataLoader
from transformers import BertTokenizerFast
from cnn_dailymail_dataset import CNNDailyMailDataset
from socket import gethostname
import evaluate
import torch
import train_utils
import os
import glob
import wandb

wandb.login()

# Output cuda/cpu
device = "cuda" if torch.cuda.is_available() else "cpu"
print(f'==============')
print(f"RUNNING ON {device.upper()}")
print(f'==============')

# Global Variable Required for cloud + collab compatibility I know its bad practice
# See its in its own if statement so you can collapse it and pretend it doesnt exist
if True:
    global save_to_google_drive
    global save_to_google_drive_always
    global save_to_google_drive_at_end
    global SAVE_TO_GOOGLE_DRIVE_OVERRIDE

    SAVE_TO_GOOGLE_DRIVE_OVERRIDE = False
    save_to_google_drive = False
    save_to_google_drive_always = False
    save_to_google_drive_at_end = True


# More compatibility stuff
host_name = gethostname()
if host_name != "IkboAcphg1":
    save_to_google_drive = True
try:
    import google.colab
    save_to_google_drive = True
except Exception:
    pass

if save_to_google_drive and SAVE_TO_GOOGLE_DRIVE_OVERRIDE:
    from pydrive.auth import GoogleAuth
    from pydrive.drive import GoogleDrive

    gauth = GoogleAuth()
    gauth.LocalWebserverAuth() # client_secrets.json need to be in the same directory as the script
    drive = GoogleDrive(gauth)
    print("Saving to Google Drive!")
    save_to_google_drive = True
else:
    save_to_google_drive = False
    print("Not Saving to Google Drive!")


# Program Config
verbose = 1
add_time_to_model_name = False

# Training config
num_epochs = 10
num_k_folds = 10  # For Cross-Validating to assess model performance
batch_size = 8
gradient_accumulation_steps = 1

# Dataset config
train_split_percentage = 1  # percentage of SPLIT - DATASET IS PRE SPLIT REMEMBER THIS
validate_split_percentage = 10  # its not comp3217
test_split_percentage = 10

use_lightning = True
use_fp16 = True
mixed_precision=16

# Print Configurations if verbose
def output_config():
    print(f"Program configuration:")
    print(f"Verbose Level: {verbose}")
    print(f"Adding time to model output: {add_time_to_model_name}\n")

    print(f"Dataset configuration:")
    print(f"Train Split Percentage: {train_split_percentage}")
    print(f"Validation Split Percentage: {validate_split_percentage}")
    print(f"Test Split Percentage: {test_split_percentage}\n")

    print(f"Training configuration:")
    print(f"Number of training epochs: {num_epochs}")
    print(f"Number of k-folds: {num_k_folds}")
    print(f"Batch size: {batch_size}")
    # TODO Implement
    #print(f"Gradient accumulation steps: {gradient_accumulation_steps}")
    #print(f"Effective Batch Size: {gradient_accumulation_steps * batch_size}\n")
    print(f"Mixed Precision: {mixed_precision}")

    print(f"Using Lightning: {use_lightning}")


if verbose > 0:
    output_config()

# Load Bert Tokenizer Fast
tokenizer = BertTokenizerFast.from_pretrained('bert-base-uncased')

# Load Datasets
train_dataset = CNNDailyMailDataset(tokenizer=tokenizer, split_type='train', split_percentage=train_split_percentage,
                                    verbose=verbose)
validation_dataset = CNNDailyMailDataset(tokenizer=tokenizer, split_type='validation',
                                         split_percentage=validate_split_percentage, verbose=verbose)
test_dataset = CNNDailyMailDataset(tokenizer=tokenizer, split_type='test', split_percentage=test_split_percentage,
                                   verbose=verbose)

# Load Datasets into data-loaders
test_loader = DataLoader(test_dataset, batch_size=batch_size)

# Define Model Object
models = [BertSingleDense(),BertDoubleDense(), BertBiLSTM()]
# Define KFold Object, set to None if not cross validating
cross_validation_k_fold = KFold(n_splits=num_k_folds) if num_k_folds > 0 else None
# Define Optimizer (AdamW) - Filters to only optimize params that are not frozen (i.e. not bert)
# Define loss function object
criterion = nn.NLLLoss()

train = False
num_cpus = os.cpu_count()
available_gpus = [torch.cuda.device(i) for i in range(torch.cuda.device_count())]
output_config()
wandb_logger = WandbLogger()
for model in models:
    if train:
        optimizer = AdamW(filter(lambda p: p.requires_grad, model.parameters()), lr=0.05)
        if not use_lightning:
            model.to(device)
            # Load Old Rouge scorer
            rouge_score = evaluate.load("rouge")
            train_utils.train_model(
                model=model,
                num_epochs=10,
                train_dataset=train_dataset,
                validation_dataset=validation_dataset,
                add_time_to_model_name=False,
                criterion=criterion,
                optimizer=optimizer,
                scorer=rouge_score,
                evaluate=True,
                verbose=1,
                save_after_epoch=True,
                save_best=True,
                device=device,
                tokenizer=tokenizer,
                batch_size=batch_size
            )  # model_name="BertSingleDense" Omitted to test custom BERT wrappers + autoname - also means this is technically
            # more generalized code
        else:
            train_loader = DataLoader(train_dataset, batch_size=batch_size)
            validation_loader = DataLoader(validation_dataset, batch_size=batch_size)
            print(available_gpus)
            model = BertLightning(model)
            print(f"Available GPUs: {len(available_gpus)}")
            logger = CSVLogger("logs", name=model.name+"Logger")
            trainer = Trainer(logger=wandb_logger)
            loss_checkpoint_callback = ModelCheckpoint(
                monitor='val_loss',
                dirpath='Models/',
                filename='{epoch}-{val_loss:.2f}-{rouge:.2f}',
                save_top_k=2,
                mode='min'
            )
            '''
            rogue2_checkpoint_callback = ModelCheckpoint(
                monitor='val_rouge2_fmeasure',
                dirpath='Models/',
                filename='{epoch}-{val_loss:.2f}-{rouge:.2f}',
                save_top_k=2,
                mode='max'
            )
            rogue_checkpoint_callback = ModelCheckpoint(
                monitor='val_rouge1_fmeasure',
                dirpath='Models/',
                filename='{epoch}-{val_loss:.2f}-{rouge:.2f}',
                save_top_k=2,
                mode='max'
            )
            '''

            print(f"Training {model.name}")
            trainer = Trainer(
                max_epochs=num_epochs,
                devices=len(available_gpus),
                accelerator="auto",
                precision="16",
                callbacks=[loss_checkpoint_callback])
            trainer.fit(model, train_loader, validation_loader)
    else:
        checkpoint_dir = "Models/"
        checkpoints = [f for f in os.listdir(checkpoint_dir)]
        for checkpoint_path in glob.glob("Models/*.ckpt"):
            for m in models:
                try:
                    model = BertLightning.load_from_checkpoint(checkpoint_path, model=m)
                    model.beam_search_decode = True
                    wandb_logger = WandbLogger()
                    batch_size = 8  # I am a mortal on local machine
                    test_loader = DataLoader(test_dataset, batch_size=batch_size, num_workers=0)
                    trainer = Trainer(logger=wandb_logger)
                    trainer.test(model=model, dataloaders=test_loader, ckpt_path=checkpoint_path)
                except Exception as pokemon:
                    print(pokemon)
                    pass
