import os
import torch
from torch.utils.data import ConcatDataset, DataLoader, SubsetRandomSampler

from analysis_utils import analyze_scores
from string_utils import build_model_name
from tqdm import trange, tqdm
import numpy as np
from drive_utils import save_to_google_drive


def train_epoch(model, data_loader, criterion, optimizer, device='cpu', epoch_number=None):
    # Set model to train mode
    model.train()
    # Init loss tracker
    loss_total = 0.0
    # Init tqdm
    desc_string = f"Epoch {epoch_number if epoch_number is not None else 'NaN'}"
    progress_bar = tqdm(data_loader, desc=desc_string, dynamic_ncols=True)
    # Iterate over each batch
    for batch_id, batch in enumerate(progress_bar):
        # Unpack batch into inputs and outputs
        input_ids = batch["input_ids"].to(device)
        attention_mask = batch["attention_mask"].to(device)
        labels = batch["labels"].to(device)  # Shape of [batch, 512]

        # Run Forward Pass
        outputs = model(input_ids=input_ids, attention_mask=attention_mask)  # Shape of [batch, 512, output layer size]
        # output layer size here is the vocab size, for classification, its the number of classes

        # Reshape logits to ensure inference on a token level - needed for context on token level and not sentence level
        # i.e. parts of speech tagging vs sentiment analysis require each token or only sentence level context
        # task dependent - not relevant as all models are seq2seq for now
        outputs_reshaped = outputs.view(-1, outputs.size(-1))  # Shape of [batch*512, n_categories]
        labels_reshaped = labels.view(-1)
        # Calculate loss
        loss = criterion(outputs_reshaped, labels_reshaped)

        # Backpropagate and optimize weights
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        loss_value = loss.item()
        loss_total += loss_value
        mean_loss = loss_total / (batch_id + 1)
        # Updates progress bar text
        progress_bar.set_postfix({'Batch': batch_id, 'Mean Loss': mean_loss, 'Last Loss': loss_value})
    # Return mean loss across entire batch
    return loss_total / len(data_loader)


# TODO Validate dataset type - fix weird coupling
def evaluate_model(model, data_loader, scorer, tokenizer, device='cpu'):
    # Set model to eval mode
    model.eval()

    # Init score tracker
    score_totals = []

    # Disable gradients for evaluation - performance
    with torch.no_grad():
        # Init tqdm
        desc_string = "Evaluation"
        progress_bar = tqdm(data_loader, desc=desc_string)

        # Iterate over each batch
        for batch_id, batch in enumerate(progress_bar):
            # Unpack batch into inputs and outputs
            input_ids = batch["input_ids"].to(device)
            attention_mask = batch["attention_mask"].to(device)
            labels = batch["labels"].to(device)

            # Run Forward Pass
            outputs = model(input_ids=input_ids, attention_mask=attention_mask)

            # Greedy decode for now
            max_values, predicted_indices = torch.max(outputs, dim=-1)
            predicted_tokens = [[tokenizer.convert_ids_to_tokens(idx.item()) for idx in seq] for seq in predicted_indices]

            # Argmax Decode
            outputs_decoded = [tokenizer.decode(o, skip_special_tokens=True) for o in predicted_indices]
            labels_decoded = [tokenizer.decode(l, skip_special_tokens=True) for l in labels]

            # Calculate performance score
            score = scorer.compute(predictions=outputs_decoded, references=labels_decoded)
            score_totals.append(score)

            # Updates progress bar text
            progress_bar.set_postfix({f'Batch {batch_id} Score': score})

    # Return scores list of dictionaries
    return score_totals


def train_model(model, num_epochs, train_dataset, validation_dataset, criterion, optimizer, scorer,
                evaluate=True, verbose=0, model_name=None, add_time_to_model_name=False,
                model_name_appendix="", save_output_dir=None, save_after_epoch=True, save_best=True, device='cpu',
                tokenizer=None, analyze_model_scores=True, cross_validation_k_folder=None, batch_size=8):

    if tokenizer is None:
        print("No tokenizer supplied to train_model. Model will not evaluate without one!")
        evaluate = False
        analyze_model_scores = False

    # Build model name
    model_name = build_model_name(model, model_name, add_time_to_model_name, model_name_appendix)

    # Sets output directory if relevant
    if save_output_dir is None and save_after_epoch:
        save_output_dir = model_name

    # Moves to subdirectory
    if save_output_dir is not None:
        save_output_dir = "Models/"+save_output_dir

    # I will kludge this into existence if its the last thing i do
    def _train_model(train_data_loader, validation_data_loader):
        all_scores = []
        best_score = 0  # 0 and not inf/-inf in case of score minimize/maximise difference
        progress_bar = trange(num_epochs, desc=f'Training {model_name}')
        for epoch in progress_bar:
            # Train model for a single epoch, obtains a loss
            loss = train_epoch(model, train_data_loader, criterion, optimizer, device=device, epoch_number=epoch)
            mean_score = 0
            if evaluate:
                # TODO Find a way to save all scores/plot cool stuff here
                # TODO Support more than Rouge
                # Save if score better than previous best
                model_evaluate_scores = evaluate_model(model, validation_data_loader, scorer, tokenizer, device=device)
                mean_scores = {k: np.mean([x[k] for x in model_evaluate_scores]) for k in model_evaluate_scores[0].keys()}
                all_scores.append(model_evaluate_scores)

                # TODO Make way of doing this dynamically
                scorer_name = "rougeL"  # This should be rouge2
                mean_score = mean_scores[scorer_name]
                if save_best and mean_score > best_score:
                    best_score = mean_score
                    save_path = os.path.join(save_output_dir, f'{model_name}_best.pt')
                    os.makedirs(save_output_dir, exist_ok=True)
                    tqdm.write(f"\nNew best model found in epoch {epoch}!\n{scorer_name} new best mean: {best_score}\nSaving model "
                               f"weights to: {save_path}")
                    torch.save(model.state_dict(), save_path)
                    if save_to_google_drive and False:
                        save_to_google_drive(model)

                # Not sure if write or postfix is better here
                #tqdm.write(f'Training: {model_name}\nEpoch: {epoch} - Loss: {loss} - {scorer_name}: {mean_score}')
                progress_bar.set_postfix({'Training': model_name, 'Epoch': epoch, 'Loss': loss, scorer_name: mean_score})


            else:
                progress_bar.set_postfix(loss=loss, mean_score=mean_score)

            # Save the model weights after each epoch
            if save_after_epoch:
                save_path = os.path.join(save_output_dir, f'{model_name}_last.pt')
                os.makedirs(save_output_dir, exist_ok=True)
                if verbose > 0:
                    tqdm.write(f"\nSaving model weights to: {save_path}")
                torch.save(model.state_dict(), save_path)
        if evaluate:
            return all_scores
        else:
            return False

    model_scores = []
    if cross_validation_k_folder is not None:
        tqdm.write(f"You are cross validating {model_name}. This may take a while.")

        # TODO check this isnt leaking data
        # Combine val/train datasets - we will validate on test set later so may as well use extra data
        dataset_combined = ConcatDataset([train_dataset, validation_dataset])

        # k-fold
        k_folder_splits = list(cross_validation_k_folder.split()) # This refuses to run otherwise
        for fold, (train_indicies, validation_indicies) in tqdm(enumerate(k_folder_splits),
                                                                total=len(k_folder_splits), desc='Folds: '):
            # Generate subsample for each fold
            train_subset_sampler = SubsetRandomSampler(train_indicies)
            validation_subset_sample = SubsetRandomSampler(validation_indicies)
            train_data_loader = DataLoader(dataset_combined, batch_size=batch_size, sampler=train_subset_sampler)
            validation_data_loader = DataLoader(dataset_combined, batch_size=batch_size, sampler=validation_subset_sample)
            fold_scores = _train_model(train_data_loader, validation_data_loader)
            model_scores.append(fold_scores)

    else:
        train_data_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        validation_data_loader = DataLoader(validation_dataset, batch_size=batch_size)
        model_scores = _train_model(train_data_loader, validation_data_loader)

    # Save best to Google Drive if enabled:
    if save_to_google_drive_at_end and save_to_google_drive:
        save_to_google_drive(model)

    # Analyze scores if they exist
    if model_scores and analyze_model_scores:
        analyze_scores(model_scores)
